import abc
import codecs
import re
import typing

from keylime.common.algorithms import Hash

# This module defines the abstraction of a Test (of JSON data)
# and several specific test classes.
# A Test can be used multiple times, even concurrently.

# Data is the type of Python data that corresponds to JSON values.
Data = typing.Union[int, float, str, bool, typing.Tuple["Data", ...], typing.Mapping[str, "Data"], None]

# Globals is a dict of variables for communication among tests.
# There is a distinct dict for each top-level use of a test.
Globals = typing.Mapping[str, Data]

# PCR_Contents maps digest name to map from PCR index to PCR value.
# Here digest name is something like 'sha256'.
# Each PCR index is a decimal string, so that this can be JSON data
PCR_Contents = typing.Mapping[str, typing.Mapping[str, int]]


class Test(metaclass=abc.ABCMeta):
    """Test is something that can examine a value and either approve it or give a reason for rejection"""

    @abc.abstractmethod
    def why_not(self, globs: Globals, subject: Data) -> str:
        """Test the given value, return empty string for pass, explanation for fail.

        The explanation is (except in deliberate exceptions) English that
        makes a sentence when placed after a noun phrase identifying the subject.
        The test can read and write in the given globs dict.
        """
        raise NotImplementedError


# type_test constructs a test of data type that is expected to pass.
# This and the following are used to check reference state for bugs.
def type_test(t) -> typing.Callable[[typing.Any], bool]:
    """Returns a lambda that tests against the given type.
    The lambda returns True on pass, raises Exception on fail."""

    def test(v: typing.Any) -> bool:
        if isinstance(v, t):
            return True
        raise Exception(f"{v!r} is a {type(v)} rather than a {t}")

    return test


def list_test(elt_test: typing.Callable[[typing.Any], bool]) -> typing.Callable[[typing.Any], bool]:
    """Return a lambda that tests for list with certain type of element"""

    def test(dat: typing.Any) -> bool:
        type_test(list)(dat)
        for elt in dat:
            elt_test(elt)
        return True

    return test


def dict_test(
    dom_test: typing.Callable[[typing.Any], bool], rng_test: typing.Callable[[typing.Any], bool]
) -> typing.Callable[[typing.Any], bool]:
    """Return a lambda that tests for dict with certain type key and value"""

    def test(dat: typing.Any) -> bool:
        type_test(dict)(dat)
        for dom, rng in dat.items():
            dom_test(dom)
            rng_test(rng)
        return True

    return test


def obj_test(**field_tests: typing.Callable[[typing.Any], bool]) -> typing.Callable[[typing.Any], bool]:
    """Return a lambda that tests for dict with string keys and a particular type for each key"""

    def test(dat: typing.Any) -> bool:
        type_test(dict)(dat)
        dom_test = type_test(str)
        for dom, rng in dat.items():
            dom_test(dom)
            if dom not in field_tests:
                continue
            rng_test = field_tests[dom]
            rng_test(rng)
        missing = set(field_tests.keys()) - set(dat.keys())
        if missing:
            raise Exception(f"{dat!r} lacks fields {missing}")
        return True

    return test


class AcceptAll(Test):
    """Every value passes this test"""

    def why_not(self, _: Globals, subject: Data) -> str:
        return ""


class RejectAll(Test):
    """No value passes this test"""

    def __init__(self, why: str):
        super().__init__()
        if not why:
            raise Exception(f"the truth value of {why!r} is false")
        self.why = why

    def why_not(self, _: Globals, subject: Data) -> str:
        return self.why


class And(Test):
    """Conjunction of given tests

    The tests are run in series, stopping as soon as one fails."""

    def __init__(self, *tests: Test):
        super().__init__()
        list(map(type_test(Test), tests))
        self.tests = tests

    def why_not(self, globs: Globals, subject: Data) -> str:
        for test in self.tests:
            reason = test.why_not(globs, subject)
            if reason:
                return reason
        return ""


class Or(Test):
    """Disjunction of given tests

    The tests are run in series, stopping as soon as one succeeds."""

    def __init__(self, *tests: Test):
        super().__init__()
        list(map(type_test(Test), tests))
        self.tests = tests

    def why_not(self, globs: Globals, subject: Data) -> str:
        if not self.tests:
            return "does not pass empty disjunction"
        reasons = []
        for test in self.tests:
            reason = test.why_not(globs, subject)
            if not reason:
                return ""
            reasons.append(reason)
        return "[" + ", ".join(reasons) + "]"


class Dispatcher(Test):
    """Apply a specific test for each key tuple.

    This kind of test applies when the subject is a dict and
    it is desired to apply a different test depending on
    the value(s) of one or more entries."""

    def __init__(self, key_names: typing.Tuple[str, ...]):
        """Initialize a Dispatcher Test.

        key_names identifies the subject dict entries that determine
        which subsidiary test to apply."""
        super().__init__()
        if len(key_names) < 1:
            raise Exception("Dispatcher given empty list of key names")
        list(map(type_test(str), key_names))
        self.key_names = key_names
        self.tests = {}

    def set(self, key_vals: typing.Tuple[typing.Union[int, str], ...], test: Test) -> None:
        """Set the test for the given value tuple"""
        if len(key_vals) != len(self.key_names):
            raise Exception(f"{key_vals!a} does not match length of {self.key_names}")
        if key_vals in self.tests:
            raise Exception(f"multiple tests for {key_vals!a}")
        self.tests[key_vals] = test

    def why_not(self, globs: Globals, subject: Data) -> str:
        if not isinstance(subject, dict):
            return "is not a dict"
        key_vals = tuple()
        for kn in self.key_names:
            if kn not in subject:
                return f"has no {kn}"
            key_vals += (subject[kn],)
        test = self.tests.get(key_vals)
        if test is None:
            return f"has unexpected {self.key_names} combination {key_vals}"
        return test.why_not(globs, subject)


class FieldTest(Test):
    """Applies given test to field having given name"""

    def __init__(self, field_name: str, field_test: Test, show_name: bool = True):
        super().__init__()
        type_test(str)(field_name)
        type_test(Test)(field_test)
        self.field_name = field_name
        self.field_test = field_test
        self.show_name = show_name

    def why_not(self, globs: Globals, subject: Data) -> str:
        if not isinstance(subject, dict):
            return "is not a dict"
        if self.field_name not in subject:
            return f"has no {self.field_name!a} field"
        reason = self.field_test.why_not(globs, subject[self.field_name])
        if reason and self.show_name:
            return self.field_name + " " + reason
        return reason


class FieldsTest(And):
    """Tests a collection of fields"""

    def __init__(self, **fields: Test):
        tests = [FieldTest(field_name, field_test) for field_name, field_test in fields.items()]
        super().__init__(*tests)


class IterateTest(Test):
    """Applies a test to every member of a list"""

    def __init__(self, elt_test: Test, show_elt: bool = False):
        super().__init__()
        self.elt_test = elt_test
        self.show_elt = show_elt

    def why_not(self, globs: Globals, subject: Data) -> str:
        if not isinstance(subject, list):
            return "is not a list"
        for idx, elt in enumerate(subject):
            reason = self.elt_test.why_not(globs, elt)
            if not reason:
                continue
            if self.show_elt:
                return f"{elt!a} " + reason
            return f"[{idx}] " + reason
        return ""


class TupleTest(Test):
    """Applies a sequence of tests to a sequence of values

    The tests are run in series, stopping as soon as one fails"""

    def __init__(self, *member_tests: Test, pad: bool = False):
        super().__init__()
        list(map(type_test(Test), member_tests))
        self.member_tests = member_tests
        self.pad = pad

    def why_not(self, globs: Globals, subject: Data) -> str:
        if not isinstance(subject, list):
            return "is not a list"
        subject_len = len(subject)
        test_len = len(self.member_tests)
        if subject_len > test_len:
            return f" is longer ({subject_len}) than the applicable tests ({test_len})"
        if (subject_len < test_len) and not self.pad:
            return f" is shorter ({subject_len}) than the applicable tests ({test_len})"
        for idx, test in enumerate(self.member_tests):
            subject_elt = subject[idx] if idx < subject_len else None
            reason = test.why_not(globs, subject_elt)
            if reason:
                return f"[{idx}] " + reason
        return ""


class DelayedField(Test):
    """Remembers a field value for later testing"""

    def __init__(self, delayer: "DelayToFields", field_name: str):
        super().__init__()
        self.delayer = delayer
        self.field_name = field_name

    def why_not(self, globs: Globals, subject: Data) -> str:
        """Add the value to the list stashed for later testing"""
        val_list = globs[self.field_name]
        if not isinstance(val_list, list):
            return f"malformed test: global {self.field_name} is not a list"
        val_list.append(subject)
        return ""


class DelayInitializer(Test):
    """A Test that initializes the globals used by a DelayToFields and reports acceptance"""

    def __init__(self, delayer: "DelayToFields"):
        super().__init__()
        self.delayer = delayer

    def why_not(self, globs: Globals, subject):
        self.delayer.initialize_globals(globs)
        return ""


class DelayToFields(Test):
    """A test to apply after stashing fields to test.

    For each field, accumulates a list of values
    in a correspondingly-named global.
    As a test, ignores the given subject and instead applies the
    configured fields_test to the record of accumulated value lists.
    """

    def __init__(self, fields_test: Test, *field_names: str):
        super().__init__()
        self.field_names = field_names
        self.fields_test = fields_test

    def initialize_globals(self, globs: Globals) -> None:
        """Initialize for a new pass over data"""
        for field_name in self.field_names:
            globs[field_name] = []

    def get_initializer(self) -> DelayInitializer:
        """Get a Test that accepts the subject and initializes the relevant globals"""
        return DelayInitializer(self)

    def get(self, field_name: str) -> DelayedField:
        """Return a Test that adds the subject to the list stashed for later evaulation"""
        if field_name not in self.field_names:
            raise Exception(f"{field_name} not in {self.field_names}")
        return DelayedField(self, field_name)

    def why_not(self, globs: Globals, subject: Data) -> str:
        """Test the stashed field values"""
        delayed = {}
        for field_name in self.field_names:
            delayed[field_name] = globs.get(field_name, None)
        return self.fields_test.why_not(globs, delayed)


class IntEqual(Test):
    """Compares with a given int"""

    def __init__(self, expected: int):
        super().__init__()
        type_test(int)(expected)
        self.expected = expected

    def why_not(self, _: Globals, subject: Data) -> str:
        if not isinstance(subject, int):
            return "is not a int"
        if subject == self.expected:
            return ""
        return f"is not {self.expected}"


class StringEqual(Test):
    """Compares with a given string"""

    def __init__(self, expected: str):
        super().__init__()
        type_test(str)(expected)
        self.expected = expected

    def why_not(self, _: Globals, subject: Data) -> str:
        if not isinstance(subject, str):
            return "is not a str"
        if subject == self.expected:
            return ""
        return f"is not {self.expected!a}"


class RegExp(Test):
    """Does a full match against a regular expression"""

    def __init__(self, pattern: str, flags=0):
        super().__init__()
        self.regexp = re.compile(pattern, flags)

    def why_not(self, _: Globals, subject: Data) -> str:
        if not isinstance(subject, str):
            return "is not a str"
        if self.regexp.fullmatch(subject):
            return ""
        return f"does not match {self.regexp.pattern}"


# hash algorithm -> hash value in hex (sans leading 0x)
Digest = typing.Mapping[str, str]


class DigestsTest(Test):
    """Tests whether subject has a digest that is in a list of good ones"""

    def __init__(self, good_digests_list: typing.Iterable[Digest]):
        """good_digests_list is a list of good {alg:hash}"""
        super().__init__()
        self.good_digests = {}
        "map from alg to set of good digests"
        for good_digests in good_digests_list:
            type_test(dict)(good_digests)
            for alg, hash_val in good_digests.items():
                if alg in self.good_digests:
                    self.good_digests[alg].add(hash_val)
                else:
                    self.good_digests[alg] = set((hash_val,))

    def why_not(self, _: Globals, subject: Data) -> str:
        if not isinstance(subject, dict):
            return "is not a dict"
        if "Digests" not in subject:
            return "has no Digests"
        digest_list = subject["Digests"]
        if not isinstance(digest_list, list):
            return "Digests is not a list"
        for idx, subject_digest in enumerate(digest_list):
            if not isinstance(subject_digest, dict):
                return f"Digests[{idx}] is {subject_digest!r}, not a dict"
            if "AlgorithmId" not in subject_digest:
                return f"digest {idx} has no AlgorithmId"
            alg = subject_digest["AlgorithmId"]
            if not isinstance(alg, str):
                return f"Digests[{idx}].AlgorithmId is {alg!r}, not a str"
            if "Digest" not in subject_digest:
                return f"digest {idx} has no Digest"
            hash_val = subject_digest["Digest"]
            if not isinstance(hash_val, str):
                return f"Digests[{idx}].Digest is {hash_val!r}, not a str"
            if alg not in self.good_digests:
                continue
            if hash_val in self.good_digests[alg]:
                return ""
        return f"has no digest approved by {self.good_digests}"


class DigestTest(DigestsTest):
    """Tests whether subject has a digest that equals a given one"""

    def __init__(self, good_digest: Digest):
        super().__init__([good_digest])


StrOrRE = typing.Union[str, typing.Pattern]


class VariableTest(Test):
    """Test whether a given variable has value passing given test"""

    def __init__(self, variable_name: str, unicode_name: StrOrRE, data_test: Test):
        """variable_name and unicode_name are as in the parsed event; data_test applies to VariableData"""
        super().__init__()
        self.variable_name = variable_name
        # pylint: disable=isinstance-second-argument-not-valid-type
        if not isinstance(unicode_name, (str, typing.Pattern)):
            # pylint: enable=isinstance-second-argument-not-valid-type
            raise Exception(f"unicode_name={unicode_name!r} is neither a str nor an re.Pattern")
        self.unicode_name = unicode_name
        self.data_test = data_test

    def why_not(self, globs: Globals, subject: Data) -> str:
        if not isinstance(subject, dict):
            return "is not a dict"
        if "Event" not in subject:
            return "has no Event field"
        evt = subject["Event"]
        if not isinstance(evt, dict):
            return "Event is not a dict"
        if "VariableName" not in evt:
            return "Event has no VariableName field"
        variable_name = evt["VariableName"]
        if variable_name != self.variable_name:
            return f"Event.VariableName is {variable_name} rather than {self.variable_name}"
        if "UnicodeName" not in evt:
            return "Event has no UnicodeName field"
        unicode_name = evt["UnicodeName"]
        if "VariableData" not in evt:
            return "Event has no VariableData field"
        if not isinstance(unicode_name, str):
            return "Event.UnicodeName is not a str"
        variable_data = evt["VariableData"]
        if isinstance(self.unicode_name, str):
            if unicode_name != self.unicode_name:
                return f"Event.UnicodeName is {unicode_name} rather than {self.unicode_name}"
        elif not self.unicode_name.fullmatch(unicode_name):
            return f"Event.UnicodeName, {unicode_name}, does not match {self.unicode_name.pattern}"
        return self.data_test.why_not(globs, variable_data)


class VariableDispatch(FieldTest):
    """Do a specific test for each variable"""

    def __init__(self):
        self.vd = Dispatcher(("VariableName", "UnicodeName"))
        super().__init__("Event", self.vd)

    def set(self, variable_name: str, unicode_name: str, data_test: Test) -> None:
        """Define the test for a specific variable"""
        self.vd.set((variable_name, unicode_name), FieldTest("VariableData", data_test))


# Signature has the following fields.
# - SignatureOwner, value is a string UUID
# - SignatureData, value is a hex string without leading 0x


Signature = typing.Mapping[str, str]


class SignatureTest(And):
    """Compares to a particular signature"""

    def __init__(self, owner: str, data: str):
        """owner is SignatureOwner, data is SignatureData"""
        super().__init__(FieldTest("SignatureOwner", StringEqual(owner)), FieldTest("SignatureData", StringEqual(data)))


class SignatureSetMember(Or):
    """Tests for membership in the given list of signatures"""

    def __init__(self, sigs: typing.Iterable[Signature]):
        tests = [SignatureTest(sig["SignatureOwner"], sig["SignatureData"]) for sig in sigs]
        super().__init__(*tests)


class KeySubset(IterateTest):
    def __init__(self, sig_type: str, keys: typing.Iterable[typing.Mapping[str, str]]):
        super().__init__(
            And(
                FieldTest("SignatureType", StringEqual(sig_type)),
                FieldTest("Keys", IterateTest(SignatureSetMember(keys))),
            )
        )


class FieldsMismatchError(Exception):
    """Represents a mismatch between expected and actual sets of field names."""

    def __init__(self, expected, actual):
        """Constructor."""
        super().__init__(expected, actual)
        type_test(set)(expected)
        type_test(set)(actual)
        list(map(type_test(str), expected))
        list(map(type_test(str), actual))
        self.expected = expected
        self.actual = actual

    def __str__(self):
        return f"expected fields {self.expected} but got {self.actual}"


class SupersetOfDicts(Test):
    """Tests that the subject is a list of dicts with at least certain members

    All dicts must have the same field names"""

    @staticmethod
    def dict_to_tuple(it: dict, field_names: typing.Tuple[str]) -> typing.Tuple:
        actual_keys = set(it.keys())
        expected_keys = set(field_names)
        if actual_keys != expected_keys:
            raise FieldsMismatchError(expected_keys, actual_keys)
        return tuple(it.get(field_name) for field_name in field_names)

    def __init__(self, reqs: typing.Iterable[dict], field_names: typing.Tuple[str]):
        list(map(type_test(dict), reqs))
        type_test(tuple)(field_names)
        list(map(type_test(str), field_names))
        self.field_names = field_names
        self.reqs = {SupersetOfDicts.dict_to_tuple(req, field_names) for req in reqs}

    def why_not(self, globs: Globals, subject: Data) -> str:
        if not isinstance(subject, list):
            return "is not a list"
        actual = set()
        for elt in subject:
            if not isinstance(elt, dict):
                return f"member {elt} is not a dict"
            try:
                tup = SupersetOfDicts.dict_to_tuple(elt, self.field_names)
            except FieldsMismatchError:
                return f"member {elt!r} does not have the right set of field names {self.field_names}"
            actual.add(tup)
        missing = self.reqs - actual
        if not missing:
            return ""
        return f"lacks {self.field_names} combinations {missing}"


class KeySuperset(TupleTest):
    """Tests that there is one Keys dict containing at least certain members"""

    def __init__(self, sig_type: str, keys: typing.Iterable[Signature]):
        super().__init__(
            And(
                FieldTest("SignatureType", StringEqual(sig_type)),
                FieldTest("Keys", SupersetOfDicts(keys, ("SignatureOwner", "SignatureData"))),
            )
        )


class OnceTest(Test):
    """Tests that only works once"""

    def __init__(self, test: Test):
        self.executed = False
        self.test = test

    def why_not(self, globs: Globals, subject: Data) -> str:
        if self.executed:
            return "test was already run once"

        self.executed = True
        return self.test.why_not(globs, subject)


# Following tests are TCG PC Client Platform specific, but are common and reduce the use of AcceptAll


class EvSeperatorTest(Or):
    """Test for valid EV_SEPARATOR entry values"""

    def __init__(self):
        # See TCG PC Client Platform Firmware Profile (Table 9 Events)
        valid_hex_values = ["00000000", "FFFFFFFF"]
        tests = []
        for value in valid_hex_values:
            val_bytes = codecs.decode(value.encode(), "hex")
            event_test = FieldTest("Event", StringEqual(value))
            digests = {}
            for hash_alg in Hash:
                digests[str(hash_alg)] = codecs.encode(hash_alg.hash(val_bytes), "hex").decode("utf-8")
            tests.append(And(event_test, DigestTest(digests)))
        super().__init__(*tests)


class EvEfiActionTest(Test):
    """Test for valid EV_EFI_ACTION entry values"""

    _expected_strings = {
        4: ["Calling EFI Application from Boot Option", "Returning from EFI Application from Boot Option"],
        5: [
            "Exit Boot Services Invocation",
            "Exit Boot Services Returned with Failure",
            "Exit Boot Services Returned with Success",
        ],
        6: ["UEFI Debug Mode"],
    }

    def __init__(self, pcr: int):
        self.pcr = pcr
        if pcr not in self._expected_strings:
            self.test = None
            return

        tests = []
        for value in self._expected_strings[pcr]:
            event_test = FieldTest("Event", StringEqual(value))
            digests = {}
            for hash_alg in Hash:
                digests[str(hash_alg)] = codecs.encode(hash_alg.hash(value.encode()), "hex").decode("utf-8")
            tests.append(And(event_test, DigestTest(digests)))
        self.test = Or(*tests)

    def why_not(self, globs: Globals, subject: Data) -> str:
        if self.test is None:
            return f"No EV_EFI_ACTION event for {self.pcr} is expected in the spec"

        return self.test.why_not(globs, subject)



import logging
import sys

from keylime import api_version as keylime_api_version
from keylime import crypto, json, keylime_logging
from keylime.requests_client import RequestsClient

logger = keylime_logging.init_logging("registrar_client")
api_version = keylime_api_version.current_version()


def getData(registrar_ip, registrar_port, agent_id, tls_context):
    """
    Get the agent data from the registrar.

    This is called by the tenant code

    :returns: JSON structure containing the agent data
    """
    # make absolutely sure you don't ask for data that contains AIK keys unauthenticated
    if not tls_context:
        raise Exception("It is unsafe to use this interface to query AIKs without server-authenticated TLS.")

    response = None
    try:
        client = RequestsClient(f"{registrar_ip}:{registrar_port}", True, tls_context=tls_context)
        response = client.get(f"/v{api_version}/agents/{agent_id}")
        response_body = response.json()

        if response.status_code == 404:
            logger.critical(
                "Error: could not get agent %s data from Registrar Server: %s", agent_id, response.status_code
            )
            keylime_logging.log_http_response(logger, logging.CRITICAL, response_body)
            return None

        if response.status_code != 200:
            logger.critical("Error: unexpected http response code from Registrar Server: %s", response.status_code)
            keylime_logging.log_http_response(logger, logging.CRITICAL, response_body)
            return None

        # Check for all values that are consumed by other parts of Keylime
        if "results" not in response_body:
            logger.critical("Error: unexpected http response body from Registrar Server: %s", response.status_code)
            return None

        if "aik_tpm" not in response_body["results"]:
            logger.critical("Error: did not receive AIK from Registrar Server.")
            return None

        if "regcount" not in response_body["results"]:
            logger.critical("Error: did not receive regcount from Registrar Server.")
            return None

        if "ek_tpm" not in response_body["results"]:
            logger.critical("Error: did not receive EK from Registrar Server.")
            return None

        if "ip" not in response_body["results"]:
            logger.critical("Error: did not receive IP from Registrar Server.")
            return None

        if "port" not in response_body["results"]:
            logger.critical("Error: did not receive port from Registrar Server.")
            return None

        return response_body["results"]

    except AttributeError as e:
        if response and response.status_code == 503:
            logger.critical("Error: the registrar is not available at %s:%s", registrar_ip, registrar_port)
        else:
            logger.exception(e)

    except Exception as e:
        logger.exception(e)

    return None


def doRegisterAgent(
    registrar_ip, registrar_port, agent_id, ek_tpm, ekcert, aik_tpm, mtls_cert=None, contact_ip=None, contact_port=None
):
    """
    Register the agent with the registrar

    This is called by the agent code

    :returns: base64 encoded blob containing the aik_tpm name and a challenge. Is encrypted with ek_tpm.
    """

    data = {
        "ekcert": ekcert,
        "aik_tpm": aik_tpm,
    }
    if ekcert is None or ekcert == "emulator":
        data["ek_tpm"] = ek_tpm

    if mtls_cert is not None:
        data["mtls_cert"] = mtls_cert
    else:
        data["mtls_cert"] = "disabled"
        logger.error("Most actions require the agent to have mTLS enabled, but no cert was provided!")
    if contact_ip is not None:
        data["ip"] = contact_ip
    if contact_port is not None:
        data["port"] = contact_port

    response = None
    try:
        # The agent accesses the registrar without mTLS, meaning without client
        # certificate
        # TODO the registrar could be accessed using TLS, but without client
        # certificate verification. Currently it is accessed without TLS at all
        client = RequestsClient(f"{registrar_ip}:{registrar_port}", False)
        response = client.post(f"/v{api_version}/agents/{agent_id}", data=json.dumps(data))
        response_body = response.json()

        if response.status_code != 200:
            logger.error("Error: unexpected http response code from Registrar Server: %s", response.status_code)
            keylime_logging.log_http_response(logger, logging.ERROR, response_body)
            return None

        logger.info("Agent registration requested for %s", agent_id)

        if "results" not in response_body:
            logger.critical("Error: unexpected http response body from Registrar Server: %s", response.status_code)
            return None

        if "blob" not in response_body["results"]:
            logger.critical("Error: did not receive blob from Registrar Server: %s", response.status_code)
            return None

        return response_body["results"]["blob"]
    except Exception as e:
        if response and response.status_code == 503:
            logger.error("Agent cannot establish connection to registrar at %s:%s", registrar_ip, registrar_port)
            sys.exit()
        else:
            logger.exception(e)

    return None


def doActivateAgent(registrar_ip, registrar_port, agent_id, key):
    """
    Activate the agent with the registrar

    Contact the registrar to inform the agent has the derived key

    This is called by the agent code

    :returns:
    """
    data = {
        "auth_tag": crypto.do_hmac(key, agent_id),
    }

    # The agent accesses the registrar without mTLS, meaning without client
    # certificate
    # TODO the registrar could be accessed using TLS, but without client
    # certificate verification. Currently it is accessed without TLS at all
    client = RequestsClient(f"{registrar_ip}:{registrar_port}", False)
    response = client.put(
        f"/v{api_version}/agents/{agent_id}/activate",
        data=json.dumps(data),
    )
    response_body = response.json()

    if response.status_code == 200:
        logger.info("Registration activated for agent %s.", agent_id)
        return True

    logger.error("Error: unexpected http response code from Registrar Server: %s", str(response.status_code))
    keylime_logging.log_http_response(logger, logging.ERROR, response_body)
    return False


def doRegistrarDelete(registrar_ip, registrar_port, agent_id, tls_context):
    """
    Delete the given agent from the registrar.

    This is called by the tenant code

    :returns: The request response body
    """

    client = RequestsClient(f"{registrar_ip}:{registrar_port}", True, tls_context=tls_context)
    response = client.delete(f"/v{api_version}/agents/{agent_id}")
    response_body = response.json()

    if response.status_code == 200:
        logger.debug("Registrar deleted.")
    else:
        logger.warning("Status command response: %s Unexpected response from registrar.", response.status_code)
        keylime_logging.log_http_response(logger, logging.WARNING, response_body)

    return response_body


def doRegistrarList(registrar_ip, registrar_port, tls_context):
    """
    Get the list of registered agents from the registrar.

    This is called by the tenant code

    :returns: The request response body
    """
    client = RequestsClient(f"{registrar_ip}:{registrar_port}", True, tls_context=tls_context)
    response = client.get(f"/v{api_version}/agents/")
    response_body = response.json()

    if response.status_code != 200:
        logger.warning("Registrar returned: %s Unexpected response from registrar.", response.status_code)
        keylime_logging.log_http_response(logger, logging.WARNING, response_body)
        return None

    return response_body



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2021 Thore Sommer
"""

import unittest

from keylime.common.algorithms import Hash
from keylime.failure import Component, Failure
from keylime.ima import ast

# BEGIN TEST DATA

VALID_ENTRIES = {
    "ima-sig-rsa": (
        ast.ImaSig,
        "10 50873c47693cf9458e87eb4a02dd4f594f7a0c0f ima-sig sha1:1350320e5f7f51553bac8aa403489a1b135bc101 /usr/bin/dd 030202f3452d23010084c2a6cf7de1aeefa119220df0da265a7c44d34380f0d97002e7c778d09cfcf88c018e6595df3ee70eda926851f159332f852e7981a8fca1bc5e959958d06d234a0896861f13cc60da825905c8ed234df26c1deecfa816d5aa9bfb11b905e2814084a86b588be60423afada8dd0dd5a143774c6d890b64195ac42fb47ef5a9a00f0d6c80711d8e0c2b843ec38c02f60fd46bc46c7b4c329ad2dbb1b7625293703f9c739dc4c2bf0769126a2f3cb2cd031d1881cd0af64bf20fd474a993b48620f103a5c14999a2f17d60721bcc019a896b4138a688a59f50cb6cd94a4cfe3b8052e82dec025fef4feabb08c7ce412e3de850f903797e293ec27c329f57fd84e0",
    ),
    "ima-sig-ec": (
        ast.ImaSig,
        "10 06e804489a77ddab51b9ef27e17053c0e5d503bd ima-sig sha1:1cb84b12db45d7da8de58ba6744187db84082f0e /usr/bin/zmore 030202531f402500483046022100bff9c02dc7b270c83cc94bfec10eecd42831de2cdcb04f024369a14623bc3a91022100cc4d015ae932fb98d6846645ed7d1bb1afd4621ec9089bc087126f191886dd31",
    ),
    "ima-sig-missing": (
        ast.ImaSig,
        "10 5426cf3031a43f5bfca183d79950698a95a728f6 ima-sig sha256:f1125b940480d20ad841d26d5ea253edc0704b5ec1548c891edf212cb1a9365e /lib/modules/5.4.48-openpower1/kernel/drivers/usb/common/usb-common.ko ",
    ),
    "ima-buf": (
        ast.ImaBuf,
        "10 b7862dbbf1383ac6c7cca7f02d981a081aacb1f1 ima-buf sha1:6e0e6fc8a188ef4f059638949adca4d221946906 device_resume 6e616d653d544553543b757569643d43525950542d5645524954592d39656633326535623635623034343234613561386562343436636630653731332d544553543b63617061636974793d303b6d616a6f723d3235333b6d696e6f723d303b6d696e6f725f636f756e743d313b6e756d5f746172676574733d313b6163746976655f7461626c655f686173683d346565383065333365353635643336333430356634303238393436653837623365396563306335383661666639656630656436663561653762656237326431333b",
    ),
    "ima": (
        ast.Ima,
        "10 d7026dc672344d3ee372217bdbc7395947788671 ima 6f66d1d8e2fffcc12dfcb78c04b81fe5b8bbae4e /usr/bin/kmod",
    ),
    "ima-ng": (
        ast.ImaNg,
        "10 7936eb315fb4e74b99e7d461bc5c96049e1ee092 ima-ng sha1:bc026ae66d81713e4e852465e980784dc96651f8 /usr/lib/systemd/systemd",
    ),
}

INVALID_ENTRIES = {
    "invalid-mode": "10 7936eb315fb4e74b99e7d461bc5c96049e1ee092 not-ima sha1:bc026ae66d81713e4e852465e980784dc96651f8 /usr/lib/systemd/systemd",
    "invalid-template-hash": "10 I936eb315fb4e74b99e7d461bc5c96049e1ee092 ima-ng sha1:bc026ae66d81713e4e852465e980784dc96651f8 /usr/lib/systemd/systemd",
    "invalid-digest": "10 7936eb315fb4e74b99e7d461bc5c96049e1ee092 ima-ng sha1:Ic026ae66d81713e4e852465e980784dc96651f8 /usr/lib/systemd/systemd",
    "not-ima-string": "this is text not an ima entry",
}

# END TEST DATA


def _true(*_):
    return Failure(Component.DEFAULT)


AlwaysTrueValidator = ast.Validator({ast.Ima: _true, ast.ImaNg: _true, ast.ImaSig: _true, ast.ImaBuf: _true})


class TestImaAst(unittest.TestCase):
    def test_valid_entry_construction(self):
        hash_alg = Hash.SHA1
        for name, (expected_mode, data) in VALID_ENTRIES.items():
            err = None
            try:
                entry = ast.Entry(data, AlwaysTrueValidator, ima_hash_alg=hash_alg, pcr_hash_alg=hash_alg)
                self.assertTrue(entry.pcr == "10", f"Expected pcr 10 for {name}. Got: {entry.pcr}")
                self.assertTrue(
                    isinstance(entry.mode, expected_mode)  # pylint: disable=isinstance-second-argument-not-valid-type
                )
                self.assertTrue(
                    entry.ima_template_hash == hash_alg.hash(entry.mode.bytes()),
                    f"Constructed hash of {name} does not match template hash.\n Expected: {entry.ima_template_hash}.\n Got: {entry.mode.bytes()}",
                )
                self.assertTrue(not entry.invalid(), f"Entry of {name} couldn't be validated.")
            except ast.ParserError as e:
                err = e
            if err:
                self.fail(f"Constructing entry {name} failed with: {err}")

    def test_invalid_entries(self):
        for _, data in INVALID_ENTRIES.items():
            with self.assertRaises(ast.ParserError):
                ast.Entry(data)



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2017 Massachusetts Institute of Technology.
"""

import os
import sys

import setuptools
from setuptools.command.build_py import build_py


class keylime_build(build_py):
    def run(self):

        # Generate the split configuration files, if not present
        setup_dir = os.path.dirname(os.path.abspath(__file__))
        config_dir = os.path.join(setup_dir, "config")
        keylime_conf = os.path.join(setup_dir, "keylime.conf")
        if not os.path.exists(config_dir) and os.path.exists(keylime_conf):
            sys.path.append(setup_dir)
            import scripts.convert_config as convert  # pylint: disable=C0415

            os.mkdir(config_dir)
            old_config = convert.get_config([[keylime_conf]])
            templates_dir = os.path.join(setup_dir, "scripts/templates")
            config = convert.process_versions(templates_dir, old_config)
            # Empty components list makes the script to output the config for
            # all components
            convert.output([], config, templates_dir, config_dir)

        build_py.run(self)


if __name__ == "__main__":
    setuptools.setup(cmdclass={"build_py": keylime_build})



"""add mtls_cert and AK to verifier DB

Revision ID: 63c30820fdc1
Revises: 9169f80345ed
Create Date: 2022-01-16 19:33:27.122196

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "63c30820fdc1"
down_revision = "9169f80345ed"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column("verifiermain", sa.Column("ak_tpm", sa.String(500)))
    op.add_column("verifiermain", sa.Column("mtls_cert", sa.String(2084)))


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "ak_tpm")
    op.drop_column("verifiermain", "mtls_cert")



import base64
import datetime
import hashlib
import hmac
import os
import secrets
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import List, Optional

import cryptography.hazmat.primitives.asymmetric

# Crypto implementation using python cryptography package
from cryptography import exceptions, x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.x509.oid import NameOID

aes_block_size = 16


def rsa_import_pubkey(pubkey):
    """Import a public key
    We try / except this, as its possible that `pubkey` can arrive as either str or bytes.
    """
    try:
        return serialization.load_pem_public_key(pubkey, backend=default_backend())
    except Exception:
        return serialization.load_pem_public_key(pubkey.encode("utf-8"), backend=default_backend())


def rsa_import_privkey(privkey, password=None):
    """Import a private key
    We try / except this, as its possible that `privkey` can arrive as either str or bytes.
    """
    try:
        return serialization.load_pem_private_key(privkey, password, backend=default_backend())
    except Exception:
        return serialization.load_pem_private_key(privkey.encode("utf-8"), password, backend=default_backend())


def x509_import_pubkey(pubkey):
    key = x509.load_pem_x509_certificate(pubkey, backend=default_backend())
    return key.public_key()


def rsa_generate(size):
    """Generate private key"""
    private_key = generate_private_key(65537, size, default_backend())
    return private_key


def get_public_key(private_key):
    """Derive public key from private key"""
    public_key = private_key.public_key()
    return public_key


def rsa_sign(key, message, paddingt="internal"):
    """RSA sign message"""
    if paddingt == "internal":
        _padding = padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH)

    if paddingt == "default":
        _padding = padding.PKCS1v15()

    signature = key.sign(message, _padding, hashes.SHA256())
    return base64.b64encode(signature)


def rsa_verify(public_key, message, signature):
    """RSA verify message"""
    try:
        public_key.verify(
            base64.b64decode(signature),
            message,
            padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH),
            hashes.SHA256(),
        )
    except exceptions.InvalidSignature:
        return False
    except Exception as e:
        raise e
    return True


def rsa_export_pubkey(private_key):
    """export public key"""
    return private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
    )


def rsa_export_privkey(private_key, password=None):
    """export private key"""

    return private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.BestAvailableEncryption(password.encode("utf-8"))
        if password
        else serialization.NoEncryption(),
    )


def rsa_encrypt(key, message):
    """RSA encrypt message"""
    return key.encrypt(
        bytes(message),
        cryptography.hazmat.primitives.asymmetric.padding.OAEP(
            mgf=cryptography.hazmat.primitives.asymmetric.padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )


def rsa_decrypt(key, ciphertext):
    """RSA decrypt message"""
    return key.decrypt(
        ciphertext,
        cryptography.hazmat.primitives.asymmetric.padding.OAEP(
            mgf=cryptography.hazmat.primitives.asymmetric.padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )


def get_random_bytes(size):
    """Generate random bytes"""
    return secrets.token_bytes(size)


def generate_random_key(size=32):
    """Generate random key using urandom wrapper"""
    return os.urandom(size)


def strbitxor(a, b):
    a = bytearray(a)
    b = bytearray(b)
    retval = bytearray(len(b))
    for i, _ in enumerate(a):
        retval[i] = a[i] ^ b[i]
    return bytes(retval)


def kdf(password, salt):
    mykdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=bytes(salt, encoding="utf8"),
        iterations=100000,
        backend=default_backend(),
    )
    return mykdf.derive(password.encode("utf-8"))


def do_hmac(key, value):
    """Generate HMAC"""
    h = hmac.new(key, msg=None, digestmod=hashlib.sha384)
    h.update(value.encode("utf-8"))
    return h.hexdigest()


def _is_multiple_16(s):
    """
    Ensures string's length is a multple of 16
    """
    if not (len(s) % 16) == 0:
        raise Exception("Ciphertext was not a multiple of 16 in length")


def _has_iv_material(s):
    """
    Make sure enough material for IV in ciphertext
    """
    if len(s) < aes_block_size:
        raise Exception("Ciphertext did not contain enough material for an IV")


def encrypt(plaintext, key):
    """Encrypt object"""
    if plaintext is None:
        plaintext = b""
    iv = generate_random_key(aes_block_size)
    encryptor = Cipher(algorithms.AES(key), modes.GCM(iv), backend=default_backend()).encryptor()
    cipher_text = encryptor.update(plaintext) + encryptor.finalize()
    return base64.b64encode(iv + cipher_text + encryptor.tag)


def decrypt(ciphertext, key):
    """Decrypt object"""
    ciphertext = base64.b64decode(ciphertext)
    iv = ciphertext[:aes_block_size]
    tag = ciphertext[-aes_block_size:]
    ciphertext = bytes(ciphertext[aes_block_size:-aes_block_size])

    decryptor = Cipher(algorithms.AES(key), modes.GCM(iv, tag), backend=default_backend()).decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()


def generate_selfsigned_cert(name, key, valid_until, other_ips: Optional[List] = None) -> x509.Certificate:
    subject = issuer = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, name)])
    # Add common SAN for localhost
    san_names = [
        x509.DNSName("localhost"),
        x509.DNSName("localhost.localdomain"),
        x509.IPAddress(IPv4Address("127.0.0.1")),
        x509.IPAddress(IPv6Address("::1")),
    ]
    if other_ips:
        for ip in other_ips:
            san_names.append(x509.IPAddress(ip_address(ip)))
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .add_extension(x509.SubjectAlternativeName(san_names), critical=False)
        .issuer_name(issuer)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(valid_until)
        .sign(key, hashes.SHA256(), backend=default_backend())
    )
    return cert



import unittest

from keylime.common import validators


class TestValidRegex(unittest.TestCase):
    """Tests for valid_regex."""

    def test_none(self):
        """Check that None is a valid regex."""
        self.assertEqual(validators.valid_regex(None), (True, None, None))

    def test_valid(self):
        """Check a well formed regex."""
        value = validators.valid_regex(r"a.*")
        self.assertTrue(value[0])
        self.assertEqual(value[1].pattern, r"a.*")
        self.assertEqual(value[2], None)

    def test_invalid(self):
        """Check a not valid regex."""
        value = validators.valid_regex(r"a[")
        self.assertEqual(value, (False, None, "Invalid regex: unterminated character set."))


class TestValidExcludeList(unittest.TestCase):
    """Tests for valid_exclude_list."""

    def test_none(self):
        """Check that the empty list is valid."""
        self.assertEqual(validators.valid_exclude_list(None), (True, None, None))

    def test_single(self):
        """Check a single exclude list element."""
        value = validators.valid_exclude_list([r"a.*"])
        self.assertTrue(value[0])
        self.assertEqual(value[1].pattern, r"(a.*)")
        self.assertEqual(value[2], None)

    def test_multi(self):
        """Check a multiple elements exclude list."""
        value = validators.valid_exclude_list([r"a.*", r"b.*"])
        self.assertTrue(value[0])
        self.assertEqual(value[1].pattern, r"(a.*)|(b.*)")
        self.assertEqual(value[2], None)

    def test_invalid(self):
        """Check an invalid exclude list."""
        value = validators.valid_exclude_list([r"a["])
        self.assertEqual(value, (False, None, "Invalid regex: unterminated character set."))


class TestValidHex(unittest.TestCase):
    """Tests for valid_hex."""

    def test_none(self):
        """Check that None is not valid."""
        self.assertFalse(validators.valid_hex(None))

    def test_empty(self):
        """Check that the empty string is not valid."""
        self.assertFalse(validators.valid_hex(""))

    def test_valid_lower(self):
        """Check a valid lower case hexadecimal number."""
        self.assertTrue(validators.valid_hex("123abc"))

    def test_valid_upper(self):
        """Check a valid upper case hexadecimal number."""
        self.assertTrue(validators.valid_hex("123ABC"))

    def test_invalid(self):
        """Check and invalid hexadecimal number."""
        self.assertFalse(validators.valid_hex("123xyz"))


class TestValidUUID(unittest.TestCase):
    """Tests for valid_uuid."""

    def test_none(self):
        """Check that None is not valid."""
        self.assertFalse(validators.valid_uuid(None))

    def test_empty(self):
        """Check that the empty string is not valid."""
        self.assertFalse(validators.valid_uuid(""))

    def test_valid(self):
        """Check a valid UUID that mix upper and lower case."""
        self.assertTrue(validators.valid_uuid("74a93e15-da24-4ff1-ABC0-55beed02a16a"))

    def test_invalid(self):
        """Check an invalid UUID that mix upper and lower case."""
        self.assertFalse(validators.valid_uuid("some text"))


class TestValidAgentID(unittest.TestCase):
    """Tests for valid_agent_id."""

    def test_none(self):
        """Check that None is not valid."""
        self.assertFalse(validators.valid_agent_id(None))

    def test_empty(self):
        """Check that the empty string is not valid."""
        self.assertFalse(validators.valid_agent_id(""))

    def test_valid_uuid(self):
        """Check a valid UUID that mix upper and lower case."""
        self.assertTrue(validators.valid_agent_id("74a93e15-da24-4ff1-ABC0-55beed02a16a"))

    def test_valid_hostname(self):
        """Check a valid hostname that mix upper and lower case."""
        self.assertTrue(validators.valid_agent_id("my-Hostname.example.com"))

    def test_invalid(self):
        """Check an invalid user ID with non valid characters."""
        self.assertFalse(validators.valid_agent_id("rm -fr *"))


if __name__ == "__main__":
    unittest.main()



import asyncio
import base64
import configparser
import datetime
import hashlib
import http.server
import importlib
import io
import multiprocessing
import os
import platform
import shutil
import signal
import socket
import subprocess
import sys
import threading
import time
import uuid
import zipfile
from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn

import psutil
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from keylime import api_version as keylime_api_version
from keylime import (
    cmd_exec,
    config,
    crypto,
    fs_util,
    json,
    keylime_logging,
    registrar_client,
    revocation_notifier,
    secure_mount,
    user_utils,
    web_util,
)
from keylime.common import algorithms, validators
from keylime.ima import ima
from keylime.tpm.tpm2_objects import pubkey_from_tpm2b_public
from keylime.tpm.tpm_abstract import TPM_Utilities
from keylime.tpm.tpm_main import tpm

# Configure logger
logger = keylime_logging.init_logging("cloudagent")

# lock required for multithreaded operation
uvLock = threading.Lock()

# Instaniate tpm
tpm_instance = tpm(need_hw_tpm=True)


class Handler(BaseHTTPRequestHandler):
    parsed_path = ""

    def do_HEAD(self):
        """Not supported"""
        web_util.echo_json_response(self, 405, "HEAD not supported")

    def do_GET(self):
        """This method services the GET request typically from either the Tenant or the Cloud Verifier.

        Only tenant and cloudverifier uri's are supported. Both requests require a nonce parameter.
        The Cloud verifier requires an additional mask paramter.  If the uri or parameters are incorrect, a 400 response is returned.
        """

        logger.info("GET invoked from %s with uri: %s", self.client_address, self.path)
        rest_params = web_util.get_restful_params(self.path)
        if rest_params is None:
            web_util.echo_json_response(self, 405, "Not Implemented: Use /version, /keys/ or /quotes/ interfaces")
            return

        if "version" in rest_params:
            version_info = {"supported_version": keylime_api_version.current_version()}
            web_util.echo_json_response(self, 200, "Success", version_info)
            return

        if not rest_params["api_version"]:
            web_util.echo_json_response(self, 400, "API Version not supported")
            return

        if "keys" in rest_params and rest_params["keys"] == "verify":
            if self.server.K is None:
                logger.info("GET key challenge returning 400 response. bootstrap key not available")
                web_util.echo_json_response(self, 400, "Bootstrap key not yet available.")
                return
            if "challenge" not in rest_params:
                logger.info("GET key challenge returning 400 response. No challenge provided")
                web_util.echo_json_response(self, 400, "No challenge provided.")
                return

            challenge = rest_params["challenge"]
            response = {}
            response["hmac"] = crypto.do_hmac(self.server.K, challenge)
            web_util.echo_json_response(self, 200, "Success", response)
            logger.info("GET key challenge returning 200 response.")

        # If agent pubkey requested
        elif "keys" in rest_params and rest_params["keys"] == "pubkey":
            response = {}
            response["pubkey"] = self.server.publickey_exportable

            web_util.echo_json_response(self, 200, "Success", response)
            logger.info("GET pubkey returning 200 response.")
            return

        elif "quotes" in rest_params:
            nonce = rest_params.get("nonce", None)
            pcrmask = rest_params.get("mask", None)
            ima_ml_entry = rest_params.get("ima_ml_entry", "0")

            # if the query is not messed up
            if nonce is None:
                logger.warning("GET quote returning 400 response. nonce not provided as an HTTP parameter in request")
                web_util.echo_json_response(self, 400, "nonce not provided as an HTTP parameter in request")
                return

            # Sanitization assurance (for tpm.run() tasks below)
            if not (nonce.isalnum() and (pcrmask is None or validators.valid_hex(pcrmask)) and ima_ml_entry.isalnum()):
                logger.warning("GET quote returning 400 response. parameters should be strictly alphanumeric")
                web_util.echo_json_response(self, 400, "parameters should be strictly alphanumeric")
                return

            if len(nonce) > tpm_instance.MAX_NONCE_SIZE:
                logger.warning(
                    "GET quote returning 400 response. Nonce is too long (max size %i): %i",
                    tpm_instance.MAX_NONCE_SIZE,
                    len(nonce),
                )
                web_util.echo_json_response(
                    self, 400, f"Nonce is too long (max size {tpm_instance.MAX_NONCE_SIZE}): {len(nonce)}"
                )
                return

            hash_alg = tpm_instance.defaults["hash"]
            quote = tpm_instance.create_quote(nonce, self.server.publickey_exportable, pcrmask, hash_alg)
            imaMask = pcrmask

            # Allow for a partial quote response (without pubkey)
            enc_alg = tpm_instance.defaults["encrypt"]
            sign_alg = tpm_instance.defaults["sign"]

            if "partial" in rest_params and (rest_params["partial"] is None or rest_params["partial"] == "1"):
                response = {
                    "quote": quote,
                    "hash_alg": hash_alg,
                    "enc_alg": enc_alg,
                    "sign_alg": sign_alg,
                }
            else:
                response = {
                    "quote": quote,
                    "hash_alg": hash_alg,
                    "enc_alg": enc_alg,
                    "sign_alg": sign_alg,
                    "pubkey": self.server.publickey_exportable,
                }

            response["boottime"] = self.server.boottime

            # return a measurement list if available
            if TPM_Utilities.check_mask(imaMask, config.IMA_PCR):
                ima_ml_entry = int(ima_ml_entry)
                if ima_ml_entry > self.server.next_ima_ml_entry:
                    ima_ml_entry = 0
                ml, nth_entry, num_entries = ima.read_measurement_list(self.server.ima_log_file, ima_ml_entry)
                if num_entries > 0:
                    response["ima_measurement_list"] = ml
                    response["ima_measurement_list_entry"] = nth_entry
                    self.server.next_ima_ml_entry = num_entries

            # similar to how IMA log retrievals are triggered by IMA_PCR, we trigger boot logs with MEASUREDBOOT_PCRs
            # other possibilities would include adding additional data to rest_params to trigger boot log retrievals
            # generally speaking, retrieving the 15Kbytes of a boot log does not seem significant compared to the
            # potential Mbytes of an IMA measurement list.
            if TPM_Utilities.check_mask(imaMask, config.MEASUREDBOOT_PCRS[0]):
                if not self.server.tpm_log_file_data:
                    logger.warning("TPM2 event log not available: %s", config.MEASUREDBOOT_ML)
                else:
                    response["mb_measurement_list"] = self.server.tpm_log_file_data

            web_util.echo_json_response(self, 200, "Success", response)
            logger.info("GET %s quote returning 200 response.", rest_params["quotes"])
            return

        else:
            logger.warning("GET returning 400 response. uri not supported: %s", self.path)
            web_util.echo_json_response(self, 400, "uri not supported")
            return

    def do_POST(self):
        """This method services the POST request typically from either the Tenant or the Cloud Verifier.

        Only tenant and cloudverifier uri's are supported. Both requests require a nonce parameter.
        The Cloud verifier requires an additional mask parameter.  If the uri or parameters are incorrect, a 400 response is returned.
        """
        rest_params = web_util.get_restful_params(self.path)

        if rest_params is None:
            web_util.echo_json_response(self, 405, "Not Implemented: Use /keys/ or /notifications/ interface")
            return

        if not rest_params["api_version"]:
            web_util.echo_json_response(self, 400, "API Version not supported")
            return

        content_length = int(self.headers.get("Content-Length", 0))
        if content_length <= 0:
            logger.warning("POST returning 400 response, expected content in message. url: %s", self.path)
            web_util.echo_json_response(self, 400, "expected content in message")
            return

        post_body = self.rfile.read(content_length)
        try:
            json_body = json.loads(post_body)
        except Exception as e:
            logger.warning("POST returning 400 response, could not parse body data: %s", e)
            web_util.echo_json_response(self, 400, "content is invalid")
            return

        if "notifications" in rest_params:
            if rest_params["notifications"] == "revocation":
                revocation_notifier.process_revocation(
                    json_body, perform_actions, cert_path=self.server.revocation_cert_path
                )
                web_util.echo_json_response(self, 200, "Success")
            else:
                web_util.echo_json_response(self, 400, "Only /notifications/revocation is supported")
            return

        if rest_params.get("keys", None) not in ["ukey", "vkey"]:
            web_util.echo_json_response(self, 400, "Only /keys/ukey or /keys/vkey are supported")
            return

        try:
            b64_encrypted_key = json_body["encrypted_key"]
            decrypted_key = crypto.rsa_decrypt(self.server.private_key, base64.b64decode(b64_encrypted_key))
        except (ValueError, KeyError, TypeError) as e:
            logger.warning("POST returning 400 response, could not parse body data: %s", e)
            web_util.echo_json_response(self, 400, "content is invalid")
            return

        have_derived_key = False

        if rest_params["keys"] == "ukey":
            if "auth_tag" not in json_body:
                logger.warning("POST returning 400 response, U key provided without an auth_tag")
                web_util.echo_json_response(self, 400, "auth_tag is missing")
                return
            self.server.add_U(decrypted_key)
            self.server.auth_tag = json_body["auth_tag"]
            self.server.payload = json_body.get("payload", None)
            have_derived_key = self.server.attempt_decryption()
        elif rest_params["keys"] == "vkey":
            self.server.add_V(decrypted_key)
            have_derived_key = self.server.attempt_decryption()
        else:
            logger.warning("POST returning  response. uri not supported: %s", self.path)
            web_util.echo_json_response(self, 400, "uri not supported")
            return
        logger.info("POST of %s key returning 200", ("V", "U")[rest_params["keys"] == "ukey"])
        web_util.echo_json_response(self, 200, "Success")

        # no key yet, then we're done
        if not have_derived_key:
            return

        # woo hoo we have a key
        # ok lets write out the key now
        secdir = secure_mount.mount()  # confirm that storage is still securely mounted

        # clean out the secure dir of any previous info before we extract files
        if os.path.isdir(os.path.join(secdir, "unzipped")):
            shutil.rmtree(os.path.join(secdir, "unzipped"))

        # write out key file
        with open(os.path.join(secdir, self.server.enc_keyname), "w", encoding="utf-8") as f:
            f.write(base64.b64encode(self.server.K).decode())

        # stow the U value for later
        tpm_instance.write_key_nvram(self.server.final_U)

        # optionally extend a hash of they key and payload into specified PCR
        tomeasure = self.server.K

        # if we have a good key, now attempt to write out the encrypted payload
        dec_path = os.path.join(secdir, config.get("agent", "dec_payload_file"))
        enc_path = os.path.join(config.WORK_DIR, "encrypted_payload")

        dec_payload = None
        enc_payload = None
        if self.server.payload is not None:
            if not self.server.mtls_cert_enabled and not config.getboolean(
                "agent", "enable_insecure_payload", fallback=False
            ):
                logger.warning(
                    'agent mTLS is disabled, and unless "enable_insecure_payload" is set to "True", payloads cannot be deployed'
                )
                enc_payload = None
            else:
                dec_payload = crypto.decrypt(self.server.payload, bytes(self.server.K))
                enc_payload = self.server.payload

        elif os.path.exists(enc_path):
            # if no payload provided, try to decrypt one from a previous run stored in encrypted_payload
            with open(enc_path, "rb") as f:
                enc_payload = f.read()
            try:
                dec_payload = crypto.decrypt(enc_payload, self.server.K)
                logger.info("Decrypted previous payload in %s to %s", enc_path, dec_path)
            except Exception as e:
                logger.warning("Unable to decrypt previous payload %s with derived key: %s", enc_path, e)
                os.remove(enc_path)
                enc_payload = None

        # also write out encrypted payload to be decrytped next time
        if enc_payload is not None:
            with open(enc_path, "wb") as f:
                f.write(self.server.payload.encode("utf-8"))

        # deal with payload
        payload_thread = None
        if dec_payload is not None:
            tomeasure = tomeasure + dec_payload
            # see if payload is a zip
            zfio = io.BytesIO(dec_payload)
            if config.getboolean("agent", "extract_payload_zip") and zipfile.is_zipfile(zfio):
                logger.info("Decrypting and unzipping payload to %s/unzipped", secdir)
                with zipfile.ZipFile(zfio, "r") as f:
                    f.extractall(os.path.join(secdir, "unzipped"))

                # run an included script if one has been provided
                initscript = config.get("agent", "payload_script")
                if initscript != "":

                    def initthread():
                        env = os.environ.copy()
                        env["AGENT_UUID"] = self.server.agent_uuid
                        with subprocess.Popen(
                            ["/bin/bash", initscript],
                            env=env,
                            shell=False,
                            cwd=os.path.join(secdir, "unzipped"),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                        ) as proc:
                            for line in iter(proc.stdout.readline, b""):
                                logger.debug("init-output: %s", line.strip())
                            # should be a no-op as poll already told us it's done
                            proc.wait()

                    if not os.path.exists(os.path.join(secdir, "unzipped", initscript)):
                        logger.info("No payload script %s found in %s/unzipped", initscript, secdir)
                    else:
                        logger.info("Executing payload script: %s/unzipped/%s", secdir, initscript)
                        payload_thread = threading.Thread(target=initthread, daemon=True)
            else:
                logger.info("Decrypting payload to %s", dec_path)
                with open(dec_path, "wb") as f:
                    f.write(dec_payload)
            zfio.close()

        # now extend a measurement of the payload and key if there was one
        pcr = config.getint("agent", "measure_payload_pcr")
        if 0 < pcr < 24:
            logger.info("extending measurement of payload into PCR %s", pcr)
            measured = tpm_instance.hashdigest(tomeasure)
            tpm_instance.extendPCR(pcr, measured)

        if payload_thread is not None:
            payload_thread.start()

        return

    # pylint: disable=W0622
    def log_message(self, format, *args):
        return


# consider using PooledProcessMixIn
# https://github.com/muayyad-alsadi/python-PooledProcessMixIn


class CloudAgentHTTPServer(ThreadingMixIn, HTTPServer):
    """Http Server which will handle each request in a separate thread."""

    # Do not modify directly unless you acquire uvLock. Set chosen for uniqueness of contained values
    u_set = set()
    v_set = set()

    private_key = None
    private_key_path = None
    private_key_password = None
    public_key = None
    publickey_exportable = None
    cert = None
    cert_path = None
    mtls_cert_enabled = False
    trusted_ca = None
    revocation_cert_path = None
    done = threading.Event()
    auth_tag = None
    payload = None
    enc_keyname = None
    K = None
    final_U = None
    agent_uuid = None
    next_ima_ml_entry = 0  # The next IMA log offset the verifier may ask for.
    boottime = int(psutil.boot_time())

    def __init__(self, server_address, RequestHandlerClass, agent_uuid, contact_ip, ima_log_file, tpm_log_file_data):
        """Constructor overridden to provide ability to pass configuration arguments to the server"""
        # Find the locations for the U/V transport and mTLS key and certificate.
        # They are either relative to secdir (/var/lib/keylime/secure) or absolute paths.
        secdir = secure_mount.mount()

        # Get server TLS files from configuration file
        (cert_path, key_path, trusted_ca, key_password), _ = web_util.get_tls_options("agent", logger=logger)

        # read or generate the key depending on configuration
        if os.path.isfile(key_path):
            # read in private key
            logger.info("Using existing key in %s", key_path)
            with open(key_path, "rb") as f:
                private_key = crypto.rsa_import_privkey(f.read(), password=key_password)
        else:
            logger.info("Key for U/V transport and mTLS certificate not found, generating a new one")
            private_key = crypto.rsa_generate(2048)
            with open(key_path, "wb") as f:
                f.write(crypto.rsa_export_privkey(private_key, password=key_password))

        self.private_key_path = key_path
        self.private_key = private_key
        self.publickey_exportable = crypto.rsa_export_pubkey(self.private_key)

        self.mtls_cert_enabled = config.getboolean("agent", "enable_agent_mtls", fallback=False)
        if self.mtls_cert_enabled:
            if os.path.isfile(cert_path):
                logger.info("Using existing mTLS cert in %s", cert_path)
                with open(cert_path, "rb") as f:
                    cert = x509.load_pem_x509_certificate(f.read(), backend=default_backend())
            else:
                logger.info("No mTLS certificate found, generating a new one")
                agent_ips = [server_address[0]]
                if contact_ip is not None:
                    agent_ips.append(contact_ip)
                with open(cert_path, "wb") as f:
                    # By default generate a TLS certificate valid for 5 years
                    valid_util = datetime.datetime.utcnow() + datetime.timedelta(days=(360 * 5))
                    cert = crypto.generate_selfsigned_cert(agent_uuid, private_key, valid_util, agent_ips)
                    f.write(cert.public_bytes(serialization.Encoding.PEM))

            if not trusted_ca:
                logger.warning("No certificates provided in 'trusted_client_ca'")

            self.trusted_ca = trusted_ca
            self.cert_path = cert_path
            self.cert = cert.public_bytes(serialization.Encoding.PEM)
        else:
            self.tls_options = None
            self.cert_path = None
            self.cert = "disabled"
            logger.info("WARNING: mTLS disabled, Tenant and Verifier will reach out to agent via HTTP")

        self.revocation_cert_path = config.get("agent", "revocation_cert")
        if self.revocation_cert_path == "default":
            self.revocation_cert_path = os.path.join(secdir, "unzipped/RevocationNotifier-cert.crt")
        elif self.revocation_cert_path[0] != "/":
            # if it is a relative, convert to absolute in work_dir
            self.revocation_cert_path = os.path.abspath(os.path.join(config.WORK_DIR, self.revocation_cert_path))

        # attempt to get a U value from the TPM NVRAM
        nvram_u = tpm_instance.read_key_nvram()
        if nvram_u is not None:
            logger.info("Existing U loaded from TPM NVRAM")
            self.add_U(nvram_u)
        http.server.HTTPServer.__init__(self, server_address, RequestHandlerClass)
        self.enc_keyname = config.get("agent", "enc_keyname")
        self.agent_uuid = agent_uuid
        self.ima_log_file = ima_log_file
        self.tpm_log_file_data = tpm_log_file_data

    def add_U(self, u):
        """Threadsafe method for adding a U value received from the Tenant

        Do not modify u_set of v_set directly.
        """
        with uvLock:
            # be very careful printing K, U, or V as they leak in logs stored on unprotected disks
            if config.INSECURE_DEBUG:
                logger.debug("Adding U len %d data:%s", len(u), base64.b64encode(u))
            self.u_set.add(u)

    def add_V(self, v):
        """Threadsafe method for adding a V value received from the Cloud Verifier
        Do not modify u_set of v_set directly.
        """
        with uvLock:
            # be very careful printing K, U, or V as they leak in logs stored on unprotected disks
            if config.INSECURE_DEBUG:
                logger.debug("Adding V: %s", base64.b64encode(v))
            self.v_set.add(v)

    def attempt_decryption(self):
        """On reception of a U or V value, this method is called to attempt the decryption of the Cloud Init script

        At least one U and V value must be received in order to attempt encryption. Multiple U and V values are stored
        to prevent an attacker from sending U/V values to deny service.
        """
        with uvLock:
            both_u_and_v_present = False
            return_value = False
            for u in self.u_set:
                for v in self.v_set:
                    both_u_and_v_present = True
                    return_value = self.decrypt_check(u, v)
                    if return_value:
                        # reset u and v sets
                        self.u_set = set()
                        self.v_set = set()
                        return return_value
            # TODO check on whether this happens or not.  NVRAM causes trouble
            if both_u_and_v_present:
                pass
                # logger.critical("Possible attack from: " + str(handler.client_address) + ".  Both U (potentially stale from TPM NVRAM) and V present but unsuccessful in attempt to decrypt check value.")
            return return_value

    def decrypt_check(self, decrypted_U, decrypted_V):
        """Decrypt the Cloud init script with the passed U and V values.

        This method will access the received auth tag, and may fail if decoy U and V values were received.
        Do not call directly unless you acquire uvLock. Returns None if decryption unsuccessful, else returns the
        decrypted agent UUID.
        """

        if self.auth_tag is None:
            return None

        if len(decrypted_U) != len(decrypted_V):
            logger.warning("Invalid U len %d or V len %d. skipping...", len(decrypted_U), len(decrypted_V))
            return None

        candidate_key = crypto.strbitxor(decrypted_U, decrypted_V)

        # be very careful printing K, U, or V as they leak in logs stored on unprotected disks
        if config.INSECURE_DEBUG:
            logger.debug("U: %s", base64.b64encode(decrypted_U))
            logger.debug("V: %s", base64.b64encode(decrypted_V))
            logger.debug("K: %s", base64.b64encode(candidate_key))

        logger.debug("auth_tag: %s", self.auth_tag)
        ex_mac = crypto.do_hmac(candidate_key, self.agent_uuid)

        if ex_mac == self.auth_tag:
            logger.info("Successfully derived K for UUID %s", self.agent_uuid)
            self.final_U = decrypted_U
            self.K = candidate_key
            return True

        logger.error("Failed to derive K for UUID %s", self.agent_uuid)

        return False


# Execute revocation action
def perform_actions(revocation):
    # load the actions from inside the keylime module
    actionlist = config.getlist("agent", "revocation_actions")
    actionlist = [f"revocation_actions.{i}" % i for i in actionlist]

    # load actions from unzipped
    secdir = secure_mount.mount()
    action_list_path = os.path.join(secdir, "unzipped/action_list")
    if os.path.exists(action_list_path):
        with open(action_list_path, encoding="utf-8") as f:
            actionlisttxt = f.read()
        if actionlisttxt.strip() != "":
            localactions = actionlisttxt.strip().split(",")
            for action in localactions:
                if not action.startswith("local_action_"):
                    logger.warning("Invalid local action: %s. Must start with local_action_", action)
                else:
                    actionlist.append(action)

            uzpath = os.path.join(secdir, "unzipped")
            if uzpath not in sys.path:
                sys.path.append(uzpath)

    for action in actionlist:
        logger.info("Executing revocation action %s", action)
        try:
            module = importlib.import_module(action)
            execute = getattr(module, "execute")
            loop = asyncio.new_event_loop()
            loop.run_until_complete(execute(revocation))
        except Exception as e:
            logger.warning("Exception during execution of revocation action %s: %s", action, e)


def revocation_listener():
    """
    This configures and starts the revocation listener. It is designed to be started in a separate process.
    """

    if config.has_option("agent", "enable_revocation_notifications"):
        if not config.getboolean("agent", "enable_revocation_notifications"):
            return

    secdir = secure_mount.mount()

    cert_path = config.get("agent", "revocation_cert")
    if cert_path == "default":
        cert_path = os.path.join(secdir, "unzipped/RevocationNotifier-cert.crt")
    elif cert_path[0] != "/":
        # if it is a relative, convert to absolute in work_dir
        cert_path = os.path.abspath(os.path.join(config.WORK_DIR, cert_path))

    try:
        while True:
            try:
                revocation_notifier.await_notifications(perform_actions, revocation_cert_path=cert_path)
            except Exception as e:
                logger.exception(e)
                logger.warning("No connection to revocation server, retrying in 10s...")
                time.sleep(10)
    except (KeyboardInterrupt, SystemExit):
        logger.info("Stopping revocation listener...")


def main():
    logger.warning("IMPORTANT: The Python agent is deprecated and will be removed with the next major release (7.0.0)!")
    logger.warning("           Please migrate to the Rust based agent: https://github.com/keylime/rust-keylime/")

    for ML in [config.MEASUREDBOOT_ML, config.IMA_ML]:
        if not os.access(ML, os.F_OK):
            logger.warning(
                'Measurement list path %s not accessible by agent. Any attempt to instruct it to access this path - via "keylime_tenant" CLI - will result in agent process dying',
                ML,
            )

    ima_log_file = None
    if os.path.exists(config.IMA_ML):
        ima_log_file = open(config.IMA_ML, "r", encoding="utf-8")  # pylint: disable=consider-using-with

    tpm_log_file_data = None
    if os.path.exists(config.MEASUREDBOOT_ML):
        with open(config.MEASUREDBOOT_ML, "rb") as tpm_log_file:
            tpm_log_file_data = base64.b64encode(tpm_log_file.read())

    if config.get("agent", "uuid") == "dmidecode":
        if os.getuid() != 0:
            raise RuntimeError("agent_uuid is configured to use dmidecode, but current process is not running as root.")
        cmd = ["which", "dmidecode"]
        ret = cmd_exec.run(cmd, raiseOnError=False)
        if ret["code"] != 0:
            raise RuntimeError("agent_uuid is configured to use dmidecode, but it's is not found on the system.")

    # initialize the tmpfs partition to store keys if it isn't already available
    secdir = secure_mount.mount()

    # Now that operations requiring root privileges are done, drop privileges
    # if 'run_as' is available in the configuration.
    if os.getuid() == 0:
        run_as = config.get("agent", "run_as", fallback="")
        if run_as != "":
            user_utils.chown(secdir, run_as)
            user_utils.change_uidgid(run_as)
            logger.info("Dropped privileges to %s", run_as)
        else:
            logger.warning("Cannot drop privileges since 'run_as' is empty or missing in keylime.conf agent section.")

    # get params for initialization
    registrar_ip = config.get("agent", "registrar_ip")
    registrar_port = config.get("agent", "registrar_port")

    # get params for the verifier to contact the agent
    contact_ip = os.getenv("KEYLIME_AGENT_CONTACT_IP", None)
    if contact_ip is None and config.has_option("agent", "contact_ip"):
        contact_ip = config.get("agent", "contact_ip")

    contact_port = os.getenv("KEYLIME_AGENT_CONTACT_PORT", None)
    if contact_port is None and config.has_option("agent", "contact_port"):
        contact_port = config.get("agent", "contact_port", fallback="invalid")

    # change dir to working dir
    fs_util.ch_dir(config.WORK_DIR)

    # set a conservative general umask
    os.umask(0o077)

    # initialize tpm
    (ekcert, ek_tpm, aik_tpm) = tpm_instance.tpm_init(
        self_activate=False, config_pw=config.get("agent", "tpm_ownerpassword")
    )  # this tells initialize not to self activate the AIK

    # Warn if kernel version is <5.10 and another algorithm than SHA1 is used,
    # because otherwise IMA will not work
    kernel_version = tuple(platform.release().split("-")[0].split("."))
    if tuple(map(int, kernel_version)) < (5, 10, 0) and tpm_instance.defaults["hash"] != algorithms.Hash.SHA1:
        logger.warning(
            "IMA attestation only works on kernel versions <5.10 with SHA1 as hash algorithm. "
            'Even if ascii_runtime_measurements shows "%s" as the '
            "algorithm, it might be just padding zeros",
            (tpm_instance.defaults["hash"]),
        )

    if ekcert is None and tpm_instance.is_emulator():
        ekcert = "emulator"

    # now we need the UUID
    try:
        agent_uuid = config.get("agent", "uuid")
    except configparser.NoOptionError:
        agent_uuid = None
    if agent_uuid == "hash_ek":
        ek_pubkey = pubkey_from_tpm2b_public(base64.b64decode(ek_tpm))
        ek_pubkey_pem = ek_pubkey.public_bytes(
            encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        agent_uuid = hashlib.sha256(ek_pubkey_pem).hexdigest()
    elif agent_uuid == "generate" or agent_uuid is None:
        agent_uuid = str(uuid.uuid4())
    elif agent_uuid == "dmidecode":
        cmd = ["dmidecode", "-s", "system-uuid"]
        ret = cmd_exec.run(cmd)
        sys_uuid = ret["retout"][0].decode("utf-8")
        agent_uuid = sys_uuid.strip()
        try:
            uuid.UUID(agent_uuid)
        except ValueError as e:
            raise RuntimeError(  # pylint: disable=raise-missing-from
                f"The UUID returned from dmidecode is invalid: {str(e)}"
            )
    elif agent_uuid == "hostname":
        agent_uuid = socket.getfqdn()
    elif agent_uuid == "environment":
        agent_uuid = os.getenv("KEYLIME_AGENT_UUID", None)
        if agent_uuid is None:
            raise RuntimeError("Env variable KEYLIME_AGENT_UUID is empty, but agent_uuid is set to 'environment'")
    elif not validators.valid_uuid(agent_uuid):
        raise RuntimeError("The UUID is not valid")

    if not validators.valid_agent_id(agent_uuid):
        raise RuntimeError("The agent ID set via agent uuid parameter use invalid characters")

    logger.info("Agent UUID: %s", agent_uuid)

    serveraddr = (config.get("agent", "ip"), config.getint("agent", "port"))

    server = CloudAgentHTTPServer(serveraddr, Handler, agent_uuid, contact_ip, ima_log_file, tpm_log_file_data)
    if server.mtls_cert_enabled:
        context = web_util.generate_tls_context(
            server.cert_path,
            server.private_key_path,
            server.trusted_ca,
            private_key_password=server.private_key_password,
            logger=logger,
        )
        server.socket = context.wrap_socket(server.socket, server_side=True)
    else:
        if (
            not config.getboolean("agent", "enable_insecure_payload", fallback=False)
            and config.get("agent", "payload_script") != ""
        ):
            raise RuntimeError(
                "agent mTLS is disabled, while a tenant can instruct the agent to execute code on the node. "
                'In order to allow the running of the agent, "enable_insecure_payload" has to be set to "True"'
            )

    serverthread = threading.Thread(target=server.serve_forever, daemon=True)

    # register it and get back a blob
    keyblob = registrar_client.doRegisterAgent(
        registrar_ip,
        registrar_port,
        agent_uuid,
        ek_tpm,
        ekcert,
        aik_tpm,
        mtls_cert=server.cert,
        contact_ip=contact_ip,
        contact_port=contact_port,
    )

    if keyblob is None:
        tpm_instance.flush_keys()
        raise Exception("Registration failed")

    # get the ephemeral registrar key
    key = tpm_instance.activate_identity(keyblob)

    if key is None:
        tpm_instance.flush_keys()
        raise Exception("Activation failed")

    # tell the registrar server we know the key
    retval = registrar_client.doActivateAgent(registrar_ip, registrar_port, agent_uuid, key)

    if not retval:
        tpm_instance.flush_keys()
        raise Exception("Registration failed on activate")

    # Start revocation listener in a new process to not interfere with tornado
    revocation_process = multiprocessing.Process(target=revocation_listener, daemon=True)
    revocation_process.start()

    logger.info(
        "Starting Cloud Agent on %s:%s with API version %s. Use <Ctrl-C> to stop",
        serveraddr[0],
        serveraddr[1],
        keylime_api_version.current_version(),
    )
    serverthread.start()

    def shutdown_handler(*_):
        logger.info("TERM Signal received, shutting down...")
        logger.debug("Stopping revocation notifier...")
        revocation_process.terminate()
        logger.debug("Shutting down HTTP server...")
        server.shutdown()
        server.server_close()
        serverthread.join()
        logger.debug("HTTP server stopped...")
        revocation_process.join()
        logger.debug("Revocation notifier stopped...")
        secure_mount.umount()
        logger.debug("Umounting directories...")
        tpm_instance.flush_keys()
        logger.debug("Flushed keys successfully")
        sys.exit(0)

    signal.signal(signal.SIGTERM, shutdown_handler)
    signal.signal(signal.SIGQUIT, shutdown_handler)
    signal.signal(signal.SIGINT, shutdown_handler)

    # Keep the main thread alive by waiting for the server thread
    serverthread.join()



"""Add ip and port to registrar

Revision ID: f82c4252bc4f
Revises: b4d024197413
Create Date: 2021-06-25 14:38:05.498971

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f82c4252bc4f"
down_revision = "b4d024197413"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    with op.batch_alter_table("registrarmain") as batch_op:
        batch_op.add_column(sa.Column("ip", sa.String(length=15), nullable=True))
        batch_op.add_column(sa.Column("port", sa.Integer(), nullable=True))


def downgrade_registrar():
    with op.batch_alter_table("registrarmain") as batch_op:
        batch_op.drop_column("ip")
        batch_op.drop_column("port")


def upgrade_cloud_verifier():
    pass


def downgrade_cloud_verifier():
    pass



"""add_verfier_id_to_verifiermain_table

Revision ID: b4d024197413
Revises: eb869a77abd1
Create Date: 2021-05-06 18:06:47.637118

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b4d024197413"
down_revision = "eb869a77abd1"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column("verifiermain", sa.Column("verifier_id", sa.String(100)))
    op.add_column("verifiermain", sa.Column("verifier_ip", sa.String(20)))
    op.add_column("verifiermain", sa.Column("verifier_port", sa.Integer))


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "verifier_id")
    op.drop_column("verifiermain", "verifier_ip")
    op.drop_column("verifiermain", "verifier_port")



"""Migrate allowlists to dedicated table (1/2)

Revision ID: a72aec03d720
Revises: bc3b6b551b0a
Create Date: 2022-07-21 12:17:17.779159

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a72aec03d720"
down_revision = "bf48e0c4751d"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():

    # Migrate existing agent info to the allowlists table.
    conn = op.get_bind()

    res = conn.execute("SELECT agent_id, tpm_policy, allowlist FROM verifiermain")
    results = res.fetchall()
    old_policy = [{"name": r[0], "tpm_policy": r[1], "ima_policy": r[2]} for r in results]

    # get existing table metadata
    meta = sa.MetaData(bind=conn)
    meta.reflect(only=("allowlists",))
    allowlists = meta.tables["allowlists"]

    op.bulk_insert(allowlists, old_policy)

    with op.batch_alter_table("verifiermain") as batch_op:
        batch_op.add_column(
            sa.Column("ima_policy_id", sa.Integer(), sa.ForeignKey("allowlists.id", name="fk_verifiermain_allowlists"))
        )
        batch_op.drop_column("allowlist")


def downgrade_cloud_verifier():
    # Migrate existing agent info to the allowlist column in verifiermain.
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    meta.reflect(only=("verifiermain", "allowlists"))
    verifiermain = meta.tables["verifiermain"]
    allowlists = meta.tables["allowlists"]

    res = conn.execute("SELECT name, ima_policy FROM allowlists")
    results = res.fetchall()

    # Put allowlists back into the "allowlist" column, and delete from the "allowlists" database
    for name, ima_policy in results:
        conn.execute(verifiermain.update().where(verifiermain.c.agent_id == name).values(**{"allowlist": ima_policy}))
        conn.execute(allowlists.delete().where(allowlists.c.name == name))

    # Drop the foreign key table
    with op.batch_alter_table("verifiermain") as batch_op:
        batch_op.drop_column("ima_policy_id")



import sys

from keylime import config, keylime_logging, measured_boot
from keylime.ima import file_signatures, ima
from keylime.tpm.tpm_abstract import TPM_Utilities

logger = keylime_logging.init_logging("cli_opts")


class UserError(Exception):
    pass


def enforce_pcrs(tpm_policy, policy_pcrs, protected_pcrs, pcr_use):
    policy_pcrs = list(tpm_policy.keys())
    policy_pcrs.remove("mask")

    for _pcr in policy_pcrs:
        if int(_pcr) in protected_pcrs:
            logger.error(
                'WARNING: PCR %s is specified in "tpm_policy", but will in fact be used by %s. Please remove it from policy',
                _pcr,
                pcr_use,
            )
            sys.exit(1)


def get_opts_error(args):
    if args.ima_exclude and not args.allowlist:
        return True, "--exclude cannot be used without an --allowlist"
    if args.allowlist and args.allowlist_url:
        return True, "--allowlist and --allowlist-url cannot be specified at the same time"
    if args.allowlist_url and not (args.allowlist_sig or args.allowlist_sig_url or args.allowlist_checksum):
        return (
            True,
            "--allowlist-url must have either --allowlist-sig, --allowlist-sig-url or --allowlist-checksum to verifier integrity",
        )
    if args.allowlist_sig and not (args.allowlist_url or args.allowlist):
        return True, "--allowlist-sig must have either --allowlist or --allowlist-url"
    if args.allowlist_sig_url and not (args.allowlist_url or args.allowlist):
        return True, "--allowlist-sig-url must have either --allowlist or --allowlist-url"
    if args.allowlist_checksum and not (args.allowlist_url or args.allowlist):
        return True, "--allowlist-checksum must have either --allowlist or --allowlist-url"
    if args.allowlist_sig and not args.allowlist_sig_key:
        return True, "--allowlist-sig must also have --allowlist-sig-key"
    if args.allowlist_sig_url and not args.allowlist_sig_key:
        return True, "--allowlist-sig-url must also have --allowlist-sig-key"
    if args.allowlist_sig_key and not (args.allowlist_sig or args.allowlist_sig_url):
        return True, "--allowlist-sig-key must have either --allowlist-sig or --allowlist-sig-url"
    return False, None


def process_allowlist(args):
    tpm_policy = "{}"
    ima_policy_name = ""
    mb_refstate = None
    ima_sign_verification_keys = []
    allowlist = {}

    # Set up PCR values
    if "tpm_policy" in args and args["tpm_policy"] is not None:
        tpm_policy = args["tpm_policy"]

    tpm_policy = TPM_Utilities.readPolicy(tpm_policy)
    logger.info("TPM PCR Mask from policy is %s", tpm_policy["mask"])

    if len(args.get("ima_sign_verification_keys")) > 0:
        # Auto-enable IMA (or-bit mask)
        tpm_policy["mask"] = hex(int(tpm_policy["mask"], 0) | (1 << config.IMA_PCR))

        # Add all IMA file signing verification keys to a keyring
        tenant_keyring = file_signatures.ImaKeyring()
        for filename in args["ima_sign_verification_keys"]:
            pubkey, keyidv2 = file_signatures.get_pubkey_from_file(filename)
            if not pubkey:
                raise UserError(f"File '{filename}' is not a file with a key")
            tenant_keyring.add_pubkey(pubkey, keyidv2)
        ima_sign_verification_keys = tenant_keyring.to_string()

    # Read command-line path string allowlist
    al_data = None

    if "allowlist" in args and args["allowlist"] is not None:

        enforce_pcrs(tpm_policy, list(tpm_policy.keys()), [config.IMA_PCR], "IMA")

        # Auto-enable IMA (or-bit mask)
        tpm_policy["mask"] = hex(int(tpm_policy["mask"], 0) | (1 << config.IMA_PCR))

        try:
            al_data = ima.read_allowlist(
                args["allowlist"], args["allowlist_checksum"], args["allowlist_sig"], args["allowlist_sig_key"]
            )
        except Exception as ima_e:
            raise UserError(str(ima_e)) from ima_e

    # Read command-line path string IMA exclude list
    excl_data = None
    if "ima_exclude" in args and args["ima_exclude"] is not None:
        if isinstance(args["ima_exclude"], str):
            excl_data = ima.read_excllist(args["ima_exclude"])
        elif isinstance(args["ima_exclude"], list):
            excl_data = args["ima_exclude"]
        else:
            raise UserError("Invalid exclude list provided")

    # Set up IMA
    if TPM_Utilities.check_mask(tpm_policy["mask"], config.IMA_PCR):
        # Process allowlists
        al_data["excllist"] = excl_data
        allowlist = al_data

    # Store allowlist name
    if "allowlist_name" in args and args["allowlist_name"] is not None:
        ima_policy_name = args["allowlist_name"]
        # Auto-enable IMA (or-bit mask)
        tpm_policy["mask"] = hex(int(tpm_policy["mask"], 0) | (1 << config.IMA_PCR))

    # Read command-line path string TPM event log (measured boot) reference state
    mb_refstate_data = None
    if "mb_refstate" in args and args["mb_refstate"] is not None:

        enforce_pcrs(tpm_policy, list(tpm_policy.keys()), config.MEASUREDBOOT_PCRS, "measured boot")

        # Auto-enable TPM event log mesured boot (or-bit mask)
        for _pcr in config.MEASUREDBOOT_PCRS:
            tpm_policy["mask"] = hex(int(tpm_policy["mask"], 0) | (1 << _pcr))

        logger.info("TPM PCR Mask automatically modified is %s to include IMA/Event log PCRs", tpm_policy["mask"])

        if isinstance(args["mb_refstate"], str):
            if args["mb_refstate"] == "default":
                args["mb_refstate"] = config.get("tenant", "mb_refstate")
            mb_refstate_data = measured_boot.read_mb_refstate(args["mb_refstate"])
        else:
            raise UserError("Invalid measured boot reference state (intended state) provided")

    # Set up measured boot (TPM event log) reference state
    if TPM_Utilities.check_mask(tpm_policy["mask"], config.MEASUREDBOOT_PCRS[2]):
        # Process measured boot reference state
        mb_refstate = mb_refstate_data

    return tpm_policy, mb_refstate, ima_policy_name, ima_sign_verification_keys, allowlist



import argparse
import json
import sys

from ..tpm import tpm_main
from . import policies

# This main module is just for command-line based testing.
# It implements a command to do one test.
# Invoke it with `python3 -m $packagename`, for some value of
# `$packagename` that works with your `$PYTHONPATH`.


parser = argparse.ArgumentParser()
parser.add_argument("policy_name", choices=policies.get_policy_names())
parser.add_argument("refstate_file", type=argparse.FileType("rt"))
parser.add_argument("eventlog_file", type=argparse.FileType("rb"), default=sys.stdin)
args = parser.parse_args()
policy = policies.get_policy(args.policy_name)
refstate_str = args.refstate_file.read()
refstate = json.loads(refstate_str)
log_bin = args.eventlog_file.read()
tpm = tpm_main.tpm()
_, log_data = tpm.parse_binary_bootlog(log_bin)
with open("/tmp/parsed.json", "wt", encoding="utf-8") as log_data_file:
    log_data_file.write(json.dumps(log_data, indent=True))
why_not = policy.evaluate(refstate, log_data)
if why_not:
    print(why_not, file=sys.stderr)
    sys.exit(1)
print("AOK")



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2022 Red Hat, Inc.
"""

import json
import tempfile
import unittest

from keylime.cmd import convert_ima_policy
from keylime.ima import ima

ALLOWLIST = {
    "meta": {
        "version": 6,
    },
    "hashes": {
        "boot_aggregate": ["e4cb9f5709c88376b5fc3743cd88e76b9aae8f3d992d845678de5215edb31216"],
        "/lib/modules/5.4.48-openpower1/kernel/drivers/usb/common/usb-common.ko": [
            "f1125b940480d20ad841d26d5ea253edc0704b5ec1548c891edf212cb1a9365e"
        ],
        "/lib/modules/5.4.48-openpower1/kernel/drivers/gpu/drm/drm_panel_orientation_quirks.ko": [
            "cd026b58efdf66658685430ff526490d54a430a3f0066a35ac26a8acab66c55d"
        ],
        "/usr/bin/dd": ["1350320e5f7f51553bac8aa403489a1b135bc101"],
        "/usr/bin/zmore": ["1cb84b12db45d7da8de58ba6744187db84082f0e"],
        "/usr/bin/zless": ["233ad3a8e77c63a7d9a56063ec2cad1eafa58850"],
    },
    "keyrings": {
        ".ima": ["a7d52aaa18c23d2d9bb2abb4308c0eeee67387a42259f4a6b1a42257065f3d5a"],
    },
    "ima": {"dm_policy": None},
}

EXCLUDELIST = """
boot_aggregate
/lib/modules/5.4.48-openpower1/kernel/drivers/usb/common/usb-common.ko
/lib/modules/5.4.48-openpower1/kernel/drivers/gpu/drm/drm_panel_orientation_quirks.ko
/usr/bin/dd
/usr/bin/zmore
/usr/bin/zless
"""


class TestPolicyConversion(unittest.TestCase):
    """Test the policy conversion code"""

    def test_allowlist_conversion(self):
        with tempfile.NamedTemporaryFile() as tmp_allow:
            with open(tmp_allow.name, "w", encoding="utf8") as f:
                f.write(json.dumps(ALLOWLIST))
            created_ima_policy = convert_ima_policy.convert_legacy_allowlist(tmp_allow.name)
        self.assertIsNotNone(created_ima_policy["digests"], "Created IMA policy has 'digests' field")
        self.assertEqual(
            created_ima_policy["meta"]["version"], ima.IMA_POLICY_CURRENT_VERSION, "Metadata version is correct"
        )
        self.assertEqual(
            created_ima_policy["meta"]["generator"], "convert_ima_policy.py", "Metadata generator is correct"
        )
        self.assertEqual(
            created_ima_policy["digests"][
                "/lib/modules/5.4.48-openpower1/kernel/drivers/gpu/drm/drm_panel_orientation_quirks.ko"
            ][0],
            "cd026b58efdf66658685430ff526490d54a430a3f0066a35ac26a8acab66c55d",
            "Sample digest is correct",
        )

        with tempfile.NamedTemporaryFile() as tmp_exclude:
            with open(tmp_exclude.name, "w", encoding="utf8") as f:
                f.write(EXCLUDELIST)

            updated_ima_policy = convert_ima_policy.update_ima_policy(
                created_ima_policy, excludelist_path=tmp_exclude.name
            )
        self.assertIsNotNone(created_ima_policy["excludes"], "Created IMA policy has 'excludes' field")
        self.assertIn("/usr/bin/zmore", updated_ima_policy["excludes"], "Sample exclusion path is correct")



import http.server
import unittest
from unittest.mock import MagicMock, patch

import tornado.web

from keylime import json, web_util


class TestConfig(unittest.TestCase):
    @patch("keylime.web_util.config")
    def test_get_restful_params(self, _):
        """Tests if the parsing of the parameters works"""
        version_url = "/v1.0/quotes/integrity?nonce=1234567890ABCDEFHIJ&mask=0x408000&vmask=0x808000&partial=0"
        version_params = {
            "api_version": "1.0",
            "quotes": "integrity",
            "nonce": "1234567890ABCDEFHIJ",
            "mask": "0x408000",
            "vmask": "0x808000",
            "partial": "0",
        }
        self.assertEqual(web_util.get_restful_params(version_url), version_params)

        basic_url = "/version"
        basic_params = {"version": None, "api_version": 0}
        self.assertEqual(web_util.get_restful_params(basic_url), basic_params)

    def test_json_response_tornado(self):
        """Tests JSON response output for Tornado"""
        mock_handler = MagicMock(spec=tornado.web.RequestHandler)
        test_data = {"key_1": "value", "key_2": 2}
        expected_output = json.dumps(
            {
                "code": 200,
                "status": "Success",
                "results": test_data,
            }
        ).encode("utf-8")
        res = web_util.echo_json_response(mock_handler, 200, "Success", test_data)
        self.assertTrue(res)
        mock_handler.set_status.assert_called_once_with(200)
        mock_handler.set_header.assert_called_once_with("Content-Type", "application/json")
        mock_handler.write.assert_called_once_with(expected_output)
        mock_handler.finish.assert_called_once()

    def test_json_response_http_server(self):
        """Tests JSON response output for Tornado"""
        mock_handler = MagicMock(spec=http.server.BaseHTTPRequestHandler)
        mock_handler.wfile = MagicMock()
        test_data = {"key_1": "value", "key_2": 2}
        expected_output = json.dumps(
            {
                "code": 200,
                "status": "Success",
                "results": test_data,
            }
        ).encode("utf-8")
        res = web_util.echo_json_response(mock_handler, 200, "Success", test_data)
        self.assertTrue(res)
        mock_handler.send_response.assert_called_once_with(200)
        mock_handler.send_header.assert_called_once_with("Content-Type", "application/json")
        mock_handler.end_headers.assert_called_once()
        mock_handler.wfile.write.assert_called_once_with(expected_output)


if __name__ == "__main__":
    unittest.main()



"""Associate moved allowlists to agents (2/2)

Revision ID: 4329e2d14944
Revises: a72aec03d720
Create Date: 2022-08-03 10:29:14.858393

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4329e2d14944"
down_revision = "a72aec03d720"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    # get existing table metadata
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    meta.reflect(only=("verifiermain",))
    verifiermain = meta.tables["verifiermain"]

    res = conn.execute("SELECT id, name FROM allowlists")
    results = res.fetchall()

    # Update new foreign key column with associated items in the "allowlists" table
    for allowlist_id, name in results:
        conn.execute(
            verifiermain.update().where(verifiermain.c.agent_id == name).values(**{"ima_policy_id": allowlist_id})
        )


def downgrade_cloud_verifier():
    with op.batch_alter_table("verifiermain") as batch_op:
        batch_op.add_column(
            sa.Column("allowlist", sa.Text().with_variant(sa.Text(length=429400000), "mysql"), nullable=True)
        )



"""create_allowlist_table

Revision ID: eb869a77abd1
Revises: ae898986c6e9
Create Date: 2021-01-12 10:54:45.263268

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "eb869a77abd1"
down_revision = "ae898986c6e9"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.create_table(
        "allowlists",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(255), nullable=False),
        sa.Column("tpm_policy", sa.Text(), nullable=True),
        sa.Column("vtpm_policy", sa.Text(), nullable=True),
        sa.Column("ima_policy", sa.Text().with_variant(sa.Text(length=429400000), "mysql"), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name", name="uniq_allowlists0name"),
        mysql_engine="InnoDB",
        mysql_charset="UTF8",
    )


def downgrade_cloud_verifier():
    op.drop_table("allowlists")



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2017 Massachusetts Institute of Technology.
"""


import sys
import unittest
from pathlib import Path

from cryptography import exceptions as crypto_exceptions
from cryptography.hazmat.primitives.asymmetric import padding

from keylime import ca_impl_openssl

# Useful constants for the test
PACKAGE_ROOT = Path(__file__).parents[1]
CODE_ROOT = f"{PACKAGE_ROOT}/keylime/"

# Custom imports
sys.path.insert(0, CODE_ROOT)


class OpenSSL_Test(unittest.TestCase):
    def test_openssl(self):
        _ = ca_impl_openssl.mk_cacert("my ca")
        (ca_cert, ca_pk, _) = ca_impl_openssl.mk_cacert()
        cert, _ = ca_impl_openssl.mk_signed_cert(ca_cert, ca_pk, "cert", 4)

        pubkey = ca_cert.public_key()
        try:
            pubkey.verify(
                cert.signature,
                cert.tbs_certificate_bytes,
                padding.PKCS1v15(),
                cert.signature_hash_algorithm,
            )
        except crypto_exceptions.InvalidSignature:
            self.fail("Certificate signature validation failed.")

        # Make sure serial number in cert is 4.
        self.assertIs(type(cert.serial_number), int)
        self.assertEqual(cert.serial_number, 4)


if __name__ == "__main__":
    unittest.main()
